package daobeauty.interceptor;

import java.io.Serializable;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.criterion.Example;

import daobeauty.interceptor.MethodInterceptor.ReturnType;
import daobeauty.interceptor.parser.Parser;

/**
 * Created 12.09.2008
 * @author M. Serhat Cinar
 *
 */
public class MethodNameToHibernateParser {
	@SuppressWarnings("unused")
	private static final Log LOG = LogFactory.getLog(MethodNameToHibernateParser.class);
	
	public static final Pattern PATTERN_CAMELCASE = Pattern.compile("([A-Z][a-z0-9]*)");
	
	public Object create(Session session, Class domainClass, Set<String> domainProperties, String methodName, Object [] args, ReturnType returnType){
		if (methodName==null){
			return null;
		}
		
		if (methodName.equals("findByPrimaryKey") || methodName.equals("findById")){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("findByPrimaryKey or findById must have at least one argument, the primary key");
			}
			return session.get(domainClass, (Serializable) args[0]);
		}
		
		if (methodName.equals("findByExample")){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("findByExample must have at least one example argument");
			}
			return executeCriteriaByReturnType(session.createCriteria(domainClass).add(Example.create(args[0])), returnType);
		}
		
		if (methodName.equals("findAll")){
			// case public List findAll();
			if (args==null || args.length<1){
				return executeCriteriaByReturnType(session.createCriteria(domainClass), returnType);
			}
			// case public List findAll(String hqlQuery);
			else if (args[0]!=null && String.class.isAssignableFrom(args[0].getClass())){ 
				if (args.length==1){
					return executeQueryByReturnType(session.createQuery((String) args[0]), returnType);
				}
				// case public List findAll(String hqlQuery, Map valuesByName);
				else if (args.length>=2 && args[1]!=null && Map.class.isAssignableFrom(args[1].getClass())){
					final Query query = session.createQuery((String) args[0]);
					final Map<String, Object> valuesByName = (Map<String, Object>) args[1];
					for (String key:valuesByName.keySet()){
						query.setParameter(key, valuesByName.get(key));
					}
					return executeQueryByReturnType(query, returnType);
				}
				// case public List findAll(String hqlQuery, Object[] values);
				else if (args.length>=2){
					final Query query = session.createQuery((String) args[0]);
					for (int i=1; i<args.length; i++){
						query.setParameter(i-1, args[i]);
					}
					return executeQueryByReturnType(query, returnType);
				}
			}
			
			if (args[0]==null || !String.class.isAssignableFrom(args[0].getClass())){
				throw new IllegalArgumentException("findAll found with a non-string argument at first place, where a hibernate query was expected");
			}
			throw new IllegalArgumentException("findAll without valid signature found");
		}
		
		if (methodName.startsWith("findBy")){
			final String methodNameWithoutPrefix = methodName.substring(6);
			final Criteria criteria = session.createCriteria(domainClass);
			criteria.add(new Parser(methodNameWithoutPrefix, domainProperties, args).parse());
			return executeCriteriaByReturnType(criteria, returnType);
		}
		
		return emptyByReturnType(returnType);
	}
	
	private Object emptyByReturnType(ReturnType returnType){
		if (returnType==null){
			returnType = ReturnType.List;
		}
		switch(returnType){
			case List:
			default:
				return Collections.EMPTY_LIST;
			case Single:
				return null;
		}
	}
	
	private Object executeQueryByReturnType(Query query, ReturnType returnType){
		if (returnType==null){
			returnType = ReturnType.List;
		}
		switch(returnType){
			case List:
			default:
				return query.list();
			case Single:
				return query.uniqueResult();
		}
	}
	
	private Object executeCriteriaByReturnType(Criteria criteria, ReturnType returnType){
		if (returnType==null){
			returnType = ReturnType.List;
		}
		switch(returnType){
			case List:
			default:
				return criteria.list();
			case Single:
				return criteria.uniqueResult();
		}
	}
}
