package daobeauty.interceptor;

import java.lang.reflect.Method;
import java.util.HashMap;

/**
 * Created 12.09.2008
 * @author M. Serhat Cinar
 *
 */
public class MethodSignature {
	private Method method;
	
	private String name;
	private Class returnType;
	private Class[] parameterTypes;
	
	private String signature;
	
	public MethodSignature(Method method){
		this.method = method;
		returnType = method.getReturnType();
		name = method.getName();
		parameterTypes = method.getParameterTypes();
		final StringBuilder sb = new StringBuilder();
		sb.append(returnType.getName()+" "+name+" (");
		if (parameterTypes.length>0){
			for (int i=0; i<parameterTypes.length; i++){
				if (i!=0){
					sb.append(", ");
				}
				sb.append(parameterTypes[i].getName());
			}
		}
		sb.append(")");
		signature = sb.toString();
	}

	/**
	 * @return the method
	 */
	public Method getMethod() {
		return method;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return the parameterTypes
	 */
	public Class[] getParameterTypes() {
		return parameterTypes;
	}

	/**
	 * @return the returnType
	 */
	public Class getReturnType() {
		return returnType;
	}

	/**
	 * @return the signature
	 */
	public String getSignature() {
		return signature;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + ((signature == null) ? 0 : signature.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (obj == null) return false;
		if (getClass() != obj.getClass()) return false;
		final MethodSignature other = (MethodSignature) obj;
		if (signature == null) {
			if (other.signature != null) return false;
		}
		else if (!signature.equals(other.signature)) return false;
		return true;
	}
	
	private static final HashMap<Method, String> methodSignaturesByMethod = new HashMap<Method, String>();
	
	public static String toMethodSignatureString(Method method){
		String signature = methodSignaturesByMethod.get(method);
		if (signature==null){
			
			final StringBuilder sb = new StringBuilder();
			sb.append(method.getReturnType().getName()+" "+method.getName()+" (");
			final Class [] parameterTypes = method.getParameterTypes();
			if (parameterTypes.length>0){
				for (int i=0; i<parameterTypes.length; i++){
					if (i!=0){
						sb.append(", ");
					}
					sb.append(parameterTypes[i].getName());
				}
			}
			sb.append(")");
			signature = sb.toString();
			methodSignaturesByMethod.put(method, signature);
		}
		return signature;
	}
}
