package daobeauty.interceptor.parser;

import java.util.Iterator;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;

/**
 * Created 13.09.2008
 * @author M. Serhat Cinar
 *
 */
public class Parser {
	private static final Log LOG = LogFactory.getLog(Parser.class);
	public static final Pattern PATTERN_CAMELCASE = Pattern.compile("([A-Z_][a-z0-9_]*)");
	
	private Set<String> properties;
	private String token;
	private final Stack<String> tokenStack = new Stack<String>();
	private final Stack<String> propertyStack = new Stack<String>();
	private final Stack<Object> argumentStack = new Stack<Object>();
	private final Stack<Criterion> criterionStack = new Stack<Criterion>();
	
	public Parser(String input, Set<String> properties, Object [] args){
		this.properties = properties;
		if (LOG.isDebugEnabled()){
			if (properties==null){
				LOG.debug("properties: null");
			}
			else{
				final StringBuilder sb = new StringBuilder();
				final Iterator<String> propertiesIter = properties.iterator();
				while (propertiesIter.hasNext()){
					sb.append(propertiesIter.next());
					if (propertiesIter.hasNext()){
						sb.append(", ");
					}
				}
				LOG.debug("properties: "+sb.toString());
			}
		}
		if (args!=null){
			for (int i=0; i<args.length; i++){
				argumentStack.push(args[i]);
			}
		}
		if (LOG.isDebugEnabled()){
			LOG.debug("argumentStack: "+argumentStack.toString());
		}
		// get all tokens
		final Matcher matcher = PATTERN_CAMELCASE.matcher(input);
		while (matcher.find()){
			tokenStack.addFirst(matcher.group(1));
		}
		if (LOG.isDebugEnabled()){
			final Iterator<String> tokenIter = tokenStack.iterator();
			final StringBuilder sb = new StringBuilder();
			while (tokenIter.hasNext()){
				sb.append(tokenIter.next());
				if (tokenIter.hasNext()){
					sb.append(", ");
				}
			}
			LOG.debug("tokens: "+sb.toString());
		}
		token = tokenStack.pop();
	}
	
	private Object argument(String expression){
		if (LOG.isDebugEnabled()){
			LOG.debug("argument("+expression+")");
			LOG.debug("  stack: "+argumentStack.toString());
		}
		Object argument = null;
		if (!argumentStack.isEmpty()){
			argument = argumentStack.pop();
		}
		if (LOG.isDebugEnabled()){
			LOG.debug("  return: \""+String.valueOf(argument)+"\"");
		}

		return argument;
	}
	
    private String value() {
		if (LOG.isDebugEnabled()){
			LOG.debug("value(): "+token);
		}

        return token;
    }
    
    private void match(){
		if (LOG.isDebugEnabled()){
			LOG.debug("match(): "+token);
		}
        token = popToken();
    }

    private void match(String token){
		if (LOG.isDebugEnabled()){
			LOG.debug("match("+token+")");
		}

    	if (this.token.equalsIgnoreCase(token)){
    		if (tokenStack.isEmpty()){
    			this.token = null;
    		}
    		else{
    			this.token = popToken();
    		}
    	}
    	else{
    		throw new IllegalArgumentException("xxx"); // TODO
    	}
    }

    
    private void matchEof(){
		if (LOG.isDebugEnabled()){
			LOG.debug("matchEof()");
		}

    	if (!tokenStack.isEmpty()){
    		throw new IllegalArgumentException("Unexpected symbol \""+tokenStack.peek()+"\" instead of EOF");
    	}
    }
    
    public Criterion parse(){
    	try{
			if (LOG.isDebugEnabled()){
				LOG.debug("parse()");
			}
	
			property();
			
	    	matchEof();
	    	return criterionStack.pop();
    	}
    	catch(RuntimeException e){
    		logState();
    		throw e;
    	}
    }
    
    private String popProperty(){
		if (LOG.isDebugEnabled()){
			LOG.debug("popProperty()");
			LOG.debug("  before: "+propertyStack.toString());
		}
		final String property = propertyStack.pop();
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+propertyStack.toString());
			LOG.debug("  return: "+property);
		}
		return property;
    }
    
    private void pushProperty(String property){
		if (LOG.isDebugEnabled()){
			LOG.debug("pushProperty("+property+")");
			LOG.debug("  before: "+propertyStack.toString());
		}
		// hql erwartet d1.name statt d1.Name
    	propertyStack.push(property);
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+propertyStack.toString());
		}
    }
    
    private Criterion popCriterion(){
		if (LOG.isDebugEnabled()){
			LOG.debug("popCriterion()");
			LOG.debug("  before: "+criterionStack.toString());
		}
		final Criterion criterion = criterionStack.pop();
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+criterionStack.toString());
			LOG.debug("  return: "+criterion);
		}
		return criterion;
    }
    
    private void pushCriterion(Criterion criterion){
		if (LOG.isDebugEnabled()){
			LOG.debug("pushCriterion("+criterion+")");
			LOG.debug("  before: "+criterionStack.toString());
		}
		criterionStack.push(criterion);
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+criterionStack.toString());
		}
    }
    
    private String popToken(){
		if (LOG.isDebugEnabled()){
			LOG.debug("popToken()");
			LOG.debug("  before: "+tokenStack.toString());
		}
		String token;
		if (tokenStack.isEmpty()){
			token = null;
		}
		else{
			token = tokenStack.pop(); 
		}
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+tokenStack.toString());
			LOG.debug("  return: "+token);
		}
		
    	return token;
    }
    
    private void pushToken(String token){
		if (LOG.isDebugEnabled()){
			LOG.debug("pushToken("+token+")");
			LOG.debug("  before: "+tokenStack.toString());
		}
		tokenStack.push(token);
		if (LOG.isDebugEnabled()){
			LOG.debug("  after: "+tokenStack.toString());
		}
    }
    
    private void property(){
		if (LOG.isDebugEnabled()){
			LOG.debug("*property()*");
			logState();
		}
		
    	String property = value().substring(0, 1).toLowerCase()+value().substring(1);
    	match();
    	while (value()!=null && !properties.contains(property)){
	    	if (LOG.isDebugEnabled()){
				LOG.debug("loop for not found property");
			}

    		property = property+value();
	    	if (LOG.isDebugEnabled()){
				LOG.debug("property is "+property);
			}
    		match();
    	}
    	if (!properties.contains(property)){
    		throw new IllegalArgumentException("Unknown property \""+property+"\"");
    	}
    	pushProperty(property);
    	try{
    		hqlExpression();
    	}
    	catch(RollBackException e){
    		if (LOG.isDebugEnabled()){
    			LOG.debug("property(): Rollback", e);
    		}
    		System.exit(1);
    	}
    }
    
    // einem hql expression kann ein And Or folgen, dann ein property
    private void hqlExpression() throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*hqlExpression()*");
			logState();
		}
		
    	// TODO rollback exception behandeln: letzten token zurück auf stack und mit 2 token property versuchen
    	if (tokenStack.isEmpty()){
    		LOG.debug("  tokenStack is empty");
//    		if (argumentStack.isEmpty()){
//    			LOG.debug("  argumentStack is empty");
//    			return;
//    		}
    		if (propertyStack.isEmpty()){
    			return;
    		}
    		final Object argument = argument("");
    		if (argument==null){
    			pushCriterion(Restrictions.isNull(popProperty()));
    		}
    		else{
    			pushCriterion(Restrictions.eq(popProperty(), argument));
    		}
    		if (LOG.isDebugEnabled()){
    			LOG.debug("  tokenStack is empty -> eq: "+criterionStack.peek());
    		}
    		return;
    	}
    	final String value = value().toLowerCase();
    	boolean ok = false;
    	// And
    	// NameAndAddress
    	// NameLikeAndAddress
    	// NameLikeAndNotLike <- zu kompliziert
    	
    	// Like, LikeLower, LikeUpper
    	if ("like".equals(value)){
    		like("");
    		ok = true;
    	}
    	// Ilike
    	else if ("ilike".equals(value)){
    		ilike("");
    		ok = true;
    	}
    	// I
    	// ILike
    	else if("i".equals(value)){
    		i_like("");
    		ok = true;
    	}
    	// Between
    	else if ("between".equals(value)){
    		between("");
    		ok = true;
    	}
    	// Gt
    	else if ("gt".equals(value)){
    		gt("");
    		ok = true;
    	}
    	// Lt
    	else if ("lt".equals(value)){
    		lt("");
    		ok = true;
    	}
    	// Is
    	// IsNull, IsNotNull
    	else if ("is".equals(value)){
    		is();
    		ok = true;
    	}
    	// Le
    	else if ("le".equals(value)){
    		le("");
    		ok = true;
    	}
    	// Ge
    	else if ("ge".equals(value)){
    		ge("");
    		ok = true;
    	}
    	// GreaterThan, GreaterThanOrEqual
    	else if ("greater".equals(value)){
    		greater("");
    		ok = true;
    	}
    	// LessThan, LessThanOrEqual
    	else if ("less".equals(value)){
    		less("");
    		ok = true;
    	}
    	else if ("ne".equals(value)){
    		ne();
    		ok = true;
    	}
    	else if ("not".equals(value)){
    		not("");
    		ok = true;
    	}
    	else if ("and".equals(value)){
    		and();
    		ok = true;
    	}
    	else if ("or".equals(value)){
    		or();
    		ok = true;
    	}
    	else if ("isnull".equals(value)){
    		isnull();
    		ok = true;
    	}
    	else if ("isnotnull".equals(value)){
    		isnotnull();
    		ok = true;
    	}
    	
    	if (ok){
    		if ("and".equalsIgnoreCase(value())){
    			and();
    		}
    		else if ("or".equalsIgnoreCase(value())){
    			or();
    		}
    		return;
    	}
    	
    	throw new RollBackException();
    }
    /**
	 * 
	 */
	private void logState() {
		LOG.debug("state:");
		LOG.debug("  token:          "+token);
		LOG.debug("  tokenStack:     "+tokenStack.toString());
		LOG.debug("  propertyStack:  "+propertyStack.toString());
		LOG.debug("  criterionStack: "+criterionStack.toString());
		LOG.debug("  argumentStack:  "+argumentStack.toString());
	}

    private void and() throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*and()*");
			logState();
		}
    	match("and");
    	property();
    	hqlExpression();
    	pushCriterion(Restrictions.and(popCriterion(), popCriterion()));
    }

    private void or() throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*or()*");
			logState();
		}
    	match("or");
    	property();
    	hqlExpression();
    	pushCriterion(Restrictions.or(popCriterion(), popCriterion()));
    }
    
    // Like, LikeLower, LikeUpper
    private void like(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*like("+argumentErrorPrefix+")*");
			logState();
		}

		match("like");
    	if (!tokenStack.isEmpty()){
	    	boolean lower = "lower".equalsIgnoreCase(value());
	    	boolean upper = "upper".equalsIgnoreCase(value());
	    	if (lower || upper){
	    		match();
	    		pushCriterion(Restrictions.ilike(popProperty(), argument(argumentErrorPrefix+"Like"+(lower?"Lower":"Upper"))));
	    		return;
	    	}
    	}
    	pushCriterion(Restrictions.like(popProperty(), argument("Like")));
    }
    
    private void ilike(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*ilike("+argumentErrorPrefix+")*");
			logState();
		}

		match("ilike");
		pushCriterion(Restrictions.ilike(popProperty(), argument(argumentErrorPrefix+"Ilike")));
    }

    private void i_like(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*i_like("+argumentErrorPrefix+")*");
			logState();
		}

		match("i");
    	if ("like".equalsIgnoreCase(value())){
    		match();
    		pushCriterion(Restrictions.ilike(popProperty(), argument(argumentErrorPrefix+"ILike")));
    		return;
    	}
		rollback("I");
   		throw new RollBackException();
    }
    
    private void between(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*between("+argumentErrorPrefix+")*");
			logState();
		}

    	match("between");
    	pushCriterion(Restrictions.between(popProperty(), argument(argumentErrorPrefix+"Between low"), argument(argumentErrorPrefix+"Between hi")));
    }
    
    private void gt(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*gt("+argumentErrorPrefix+")*");
			logState();
		}

    	match("gt");
    	pushCriterion(Restrictions.gt(popProperty(), argument(argumentErrorPrefix+"Gt")));
    }
    
    private void ge(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*ge("+argumentErrorPrefix+")*");
			logState();
		}
    	match("ge");
    	pushCriterion(Restrictions.ge(popProperty(), argument(argumentErrorPrefix+"Ge")));
    }
    
    private void lt(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*lt("+argumentErrorPrefix+")*");
			logState();
		}
		
    	match("lt");
    	pushCriterion(Restrictions.lt(popProperty(), argument(argumentErrorPrefix+"Lt")));
    }

    private void le(String argumentErrorPrefix){
		if (LOG.isDebugEnabled()){
			LOG.debug("*le("+argumentErrorPrefix+")*");
			logState();
		}
		
    	match("le");
    	pushCriterion(Restrictions.le(popProperty(), argument(argumentErrorPrefix+"Le")));
    }
    
    private void is() throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*is()*");
			logState();
		}
		

		match("is");
    	if ("not".equalsIgnoreCase(value())){
    		is_not_null();
    		return;
    	}
    	else if ("null".equalsIgnoreCase(value())){
    		is_null();
    		return;
    	}
    	rollback("Is");
    	throw new RollBackException();
    }
    
    private void is_null(){
		if (LOG.isDebugEnabled()){
			LOG.debug("*is_null()*");
			logState();
		}

		match("null");
		pushCriterion(Restrictions.isNull(popProperty()));
    }
    
    private void is_not_null() throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*is_not_null()*");
			logState();
		}
		
    	match("not");
    	if ("null".equalsIgnoreCase(value())){
    		match();
    		pushCriterion(Restrictions.isNotNull(popProperty()));
    		return;
    	}
    	rollback("Not");
    	throw new RollBackException();
    }
    
    private void isnull(){
		if (LOG.isDebugEnabled()){
			LOG.debug("*isnull()*");
			logState();
		}
		match("isnull");
		pushCriterion(Restrictions.isNull(popProperty()));
    }
    
    private void isnotnull(){
		if (LOG.isDebugEnabled()){
			LOG.debug("*isnotnull()*");
			logState();
		}
		match("isnotnull");
		pushCriterion(Restrictions.isNotNull(popProperty()));
    }
    
    private void greater(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*greater("+argumentErrorPrefix+")*");
			logState();
		}
		
		match("greater");
    	if ("than".equalsIgnoreCase(value())){
    		greater_than(argumentErrorPrefix);
    		return;
    	}
    	rollback("Greater");
    	throw new RollBackException();
    }
    
    private void greater_than(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*greater_than("+argumentErrorPrefix+")*");
			logState();
		}
		
		match("than");
    	if ("or".equalsIgnoreCase(value())){
    		greater_than_or_equal(argumentErrorPrefix);
    		return;
    	}
		pushCriterion(Restrictions.gt(popProperty(), argument(argumentErrorPrefix+"GreaterThan")));
    }
    
    private void greater_than_or_equal(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*greater_than_or_equal("+argumentErrorPrefix+")*");
			logState();
		}
		
		match("or");
    	if ("equal".equalsIgnoreCase(value())){
    		match();
    		pushCriterion(Restrictions.ge(popProperty(), argument(argumentErrorPrefix+"GreaterThanOrEqual")));
    		return;
    	}
    	
		rollback("Or");
    	throw new RollBackException();
    }
    
    private void less(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*less("+argumentErrorPrefix+")*");
			logState();
		}
		
    	match("less");
    	if ("than".equalsIgnoreCase(value())){
    		less_than(argumentErrorPrefix);
    		return;
    	}
    	
    	rollback("Less");
    	throw new RollBackException();
    }
    
    private void less_than(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*less_than("+argumentErrorPrefix+")*");
			logState();
		}
		
		match("less");
    	if ("or".equalsIgnoreCase(value())){
    		less_than_or_equal(argumentErrorPrefix);
    		return;
    	}
		pushCriterion(Restrictions.lt(popProperty(), argument(argumentErrorPrefix+"LessThan")));
    }
    
    private void less_than_or_equal(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*less_than_or_equal("+argumentErrorPrefix+")*");
			logState();
		}
		
    	match("or");
    	if ("equal".equalsIgnoreCase(value())){
    		pushCriterion( Restrictions.le(popProperty(), argument(argumentErrorPrefix+"LessThanOrEqual")));
    		return;
    	}
    	rollback("Or");
    	throw new RollBackException();
    }
    
    private void ne(){
		if (LOG.isDebugEnabled()){
			LOG.debug("*ne()*");
			logState();
		}
		
    	match("ne");
    	pushCriterion(Restrictions.ne(popProperty(), argument("Ne")));
    }
    
    private void not(String argumentErrorPrefix) throws RollBackException{
		if (LOG.isDebugEnabled()){
			LOG.debug("*not("+argumentErrorPrefix+")*");
			logState();
		}

    	match("not");
    	// Not
    	// NotLike, NotILike, NotIlike, NotGreaterThan.., NotLessThan..., NotEqual (Ne), NotBetween,
		final String value = value().toLowerCase();
    	if ("like".equals(value)){
    		like(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Ilike
    	else if ("ilike".equals(value)){
    		ilike(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// I
    	// ILike
    	else if("i".equals(value)){
    		i_like(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Between
    	else if ("between".equals(value)){
    		between(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Gt
    	else if ("gt".equals(value)){
    		gt(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Lt
    	else if ("lt".equals(value)){
    		lt(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Le
    	else if ("le".equals(value)){
    		le(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// Ge
    	else if ("ge".equals(value)){
    		ge(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// GreaterThan, GreaterThanOrEqual
    	else if ("greater".equals(value)){
    		greater(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	// LessThan, LessThanOrEqual
    	else if ("less".equals(value)){
    		less(argumentErrorPrefix+"Not");
    		pushCriterion(Restrictions.not(popCriterion()));
    		return;
    	}
    	rollback("Not");
    	throw new RollBackException();
    }
    
    private void rollback(String token){
    	pushToken(this.token);
    	this.token = token;
    }
}
