package daobeauty.test;

import java.util.List;

import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;

import daobeauty.proxy.ProxyFactory;
import daobeauty.test.dao.D1Dao;
import daobeauty.test.dao.D1Delegate;
import daobeauty.test.domain.D1;

/**
 * Created 12.09.2008
 * @author M. Serhat Cinar
 *
 */
public class TestReflectionInvocationHandler {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Session session = null;
		Transaction transaction = null;
		try{
			// preparartion
			final SessionFactory sessionFactory = HibernateSessionFactoryUtil.getSessionFactory();
			session = sessionFactory.getCurrentSession();
			transaction = session.beginTransaction();
			final D1Delegate delegate = new D1Delegate();
			final ProxyFactory proxyFactory = new ProxyFactory();
			
			// creation
			final D1Dao d1dao = (D1Dao) 
				proxyFactory.generateProxy(new Class[]{D1Dao.class}, sessionFactory, D1.class, null, new Object[]{delegate});
			
			// saveOrUpdate
			System.out.println("saveOrUpdate");
			D1 test = new D1();
			test.setName("a");
			d1dao.save(test);
			
			test = new D1();
			test.setName("b");
			d1dao.save(test);
			
			test = new D1();
			test.setName("bc");
			d1dao.save(test);
			
			test.setSubject("test");
			d1dao.update(test);
			
			// by named query
			System.out.println("by named query");
			List<D1> result = d1dao.findAll();
			for (D1 d:result){
				System.out.println(d.toString());
			}
			
			// by delegate
			System.out.println("by delegate");
			D1 highestId = d1dao.findByHighestId();
			System.out.println("Highest id: "+highestId);
			
			// by example
			System.out.println("by example");
			final D1 exampleD1 = new D1();
			exampleD1.setName("a");
			result = d1dao.findByExample(exampleD1);
			for (D1 d:result){
				System.out.println(d.toString());
			}
			
			// by id
			System.out.println("by id");
			D1 byPk = d1dao.findById(new Long(1));
			System.out.println(byPk.toString());
			
			// by pk
			System.out.println("by primarykey");
			byPk = d1dao.findByPrimaryKey(new Long(2));
			System.out.println(byPk.toString());
			
			// by parsing
			System.out.println("by parsing findByNameAndSubject('a', '')");
			result = d1dao.findByNameAndSubject("a", "");
			if (result.size()<1){
				System.out.println("empty result");
			}
			else{
				for (D1 d:result){
					System.out.println(d.toString());
				}			
			}
			// by parsing
			System.out.println("by parsing findByNameAndSubject('a', null)");
			result = d1dao.findByNameAndSubject("a", null);
			if (result.size()<1){
				System.out.println("empty result");
			}
			else{
				for (D1 d:result){
					System.out.println(d.toString());
				}
			}

			// by parsing
			System.out.println("by parsing findByNameILikeAndSubjectIsNotNull('A')");
			result = d1dao.findByNameILikeAndSubjectIsNull("A");
			if (result.size()<1){
				System.out.println("empty result");
			}
			else{
				for (D1 d:result){
					System.out.println(d.toString());
				}
			}

			// by parsing
			System.out.println("by parsing findByFirstAndLastName('a')");
			result = d1dao.findByFirstAndLastName("a");
			if (result.size()<1){
				System.out.println("empty result");
			}
			else{
				for (D1 d:result){
					System.out.println(d.toString());
				}
			}

			
			transaction.commit();
		}
		catch(Throwable t){
			try{
				transaction.rollback();
			}
			catch(Throwable tt){}
			t.printStackTrace();
		}
		finally{
			try{
				session.close();
			}
			catch(Throwable t){}
		}
	}
}
